"""
Numba only functions
"""

def array_ravel(arr):
    """Flatten a C/F array into a 1D array without enforcing the ordering of
    the each element.

    Args
    ----
    arr: array

    Returns
    -------
    A flattened 1D array
    """
    raise NotImplementedError
